/* player.cpp -- cut down version of editor.cpp from snd project
 *    
 * Written October 2008 by David Hugo
 */


#include <allegro.h>
#include <mem.h>
#include "sndhelp.h"

extern volatile int frame_counter;
int newfile(void);
int openfile(void);
int capture(void);
int save(void);
int copy(void);
int paste(void);
int quit(void);
int about(void);
int testtone(void);
int my_button_proc   (int msg, DIALOG *d, int c);
int my_button_proc_ex(int msg, DIALOG *d, int c);
int d_waveform_proc  (int msg, DIALOG *d, int c);
int d_playback_proc  (int msg, DIALOG *d, int c);
int play(DIALOG *d);
int pause(DIALOG *d);
int stop(DIALOG *d);
void render_waveform(BITMAP *bmp, unsigned short *data, unsigned length, unsigned fg, unsigned bg);

char *buffer;
int buffer_size;

/* the first menu in the menubar */
MENU menu1[] =
{
    { "&Open \tCtrl+O",   openfile,  NULL,      0,  NULL  },
    { "&Quit \tq/Esc",        quit,  NULL,      0,  NULL  },
    { NULL,                   NULL,  NULL,      0,  NULL  }
};



/* the help menu */
MENU helpmenu[] =
{
    { "&About \tF1",     about,  NULL,      0,  NULL  },
    { "&Test tone",   testtone,  NULL,      0,  NULL  },
    { NULL,               NULL,  NULL,      0,  NULL  }
};



/* the main menu-bar */
MENU the_menu[] =
{
   
    { "&File",  NULL,    menu1,          0,      NULL  },
    //{ "&Second", NULL,   menu2,          0,      NULL  },
    { "&Help",   NULL,   helpmenu,       0,      NULL  },
    { NULL,      NULL,   NULL,           0,      NULL  }
};



/* here it comes... the big ugly dialog structure */
DIALOG editor_dialog[] =
{
    /* (dialog proc)     (x)   (y)   (w)   (h) (fg)(bg) (key) (flags)     (d1) (d2)    (dp)                   (dp2) (dp3) */
   
    /* this element just clears the screen, therefore it should come before the others */
    { d_clear_proc,        0,   0,    0,    0,   0,  0,    0,      0,       0,   0,    NULL,                   NULL, NULL  },

    /* a menu bar - note how it auto-calculates its dimension if they are not given */
    { d_menu_proc,         0,   0,    0,    0,   0,  0,    0,      0,       0,   0,    the_menu,               NULL, NULL  },
   
    /* first waveform display */
    { d_waveform_proc,    90,  30,  700,  200,   0,  0,    0,      0,       2,   0,    NULL,                   (void*)&buffer, (void*)&buffer_size },
    { d_slider_proc,      87, 235,  706,   10,   0,  0,    0,      0,     699,   0,    NULL,                   NULL, NULL  },
    
    /* a bunch of descriptive text elements */
    { d_rtext_proc,       78, 220,    0,    0,   0,  0,    0,      0,       0,   0,    (void*)"Playback",       NULL, NULL  },
    { d_rtext_proc,       78, 235,    0,    0,   0,  0,    0,      0,       0,   0,    (void*)"cursor",         NULL, NULL  },
    
    /* the quit and side buttons use our customized dialog procedure, using dp3 as callback */
    { my_button_proc,     10,  30,   70,   20,   0,  0,  'o', D_EXIT,       0,   0,    (void*)"&Open",         NULL, (void*)openfile  },        
    { my_button_proc,    630, 280,  160,   20,   0,  0,  'q', D_EXIT,       0,   0,    (void*)"&Quit",         NULL, (void*)quit  },

    /* playback buttons and nonvisual playback procedures */
    { my_button_proc_ex,  10, 140,   70,   20,   0,  0,    0, D_EXIT,      11,   0,    (void*)"Play",          NULL, (void*)play },
    { my_button_proc_ex,  10, 165,   70,   20,   0,  0,    0, D_EXIT,      11,   0,    (void*)"Pause",         NULL, (void*)pause },
    { my_button_proc_ex,  10, 190,   70,   20,   0,  0,    0, D_EXIT,      11,   0,    (void*)"Stop",          NULL, (void*)stop },
    { d_playback_proc,     0,   0,    0,    0,   0,  0,    0,      0,       2,   0,    NULL,                   (void*)&buffer, (void*)&buffer_size  },
    
    /* keyboard shortcuts */
    { d_keyboard_proc,     0,   0,    0,    0,   0,  0,    0,      0,  KEY_F1,   0,    (void *)about,          NULL, NULL  },
    { d_keyboard_proc,     0,   0,    0,    0,   0,  0,  'o'-'`',  0,       0,   0,    (void *)openfile,       NULL, NULL  },
    { d_yield_proc,        0,   0,    0,    0,   0,  0,    0,      0,       0,   0,    NULL,                   NULL, NULL  },
    { NULL,                0,   0,    0,    0,   0,  0,    0,      0,       0,   0,    NULL,                   NULL, NULL  }
};



unsigned long xtosample(unsigned x, unsigned limit, unsigned width) {
   return (unsigned)((((unsigned long long)limit)*x)/width);
}



/* Open file callback */
int openfile(void)
{
    char filename[1024] = "";
    if (file_select_ex("Open file", filename, NULL, 1023, 480, 240) != 0) {
        alert("Selected file:", "", filename, "Ok", NULL, 0, 0);
        read_buffer_from_file(filename, buffer, buffer_size);
        render_waveform((BITMAP*)(editor_dialog[2].dp), (unsigned short*)buffer, buffer_size, editor_dialog[2].fg, editor_dialog[2].bg);
    }
    return D_O_K;
}



/* Used as a menu-callback, and by the quit button. */
int quit(void)
{
    if (alert("Really Quit?", NULL, NULL, "&Yes", "&No", 'y', 'n') == 1)
        return D_CLOSE;
    else
        return D_O_K;
}



/* Our about box. */
int about(void)
{
    alert("* Sound Decompressor *",
          "",
          "A program to playback audio from files",
          "Ok", 0, 0, 0); 
    return D_O_K;
}



int testtone(void) {
    play_test_tone();
    return D_O_K;   
}



/* A custom dialog procedure, derived from d_button_proc. It intercepts
 * the D_CLOSE return of d_button_proc, and calls the function in dp3.
 */
int my_button_proc (int msg, DIALOG *d, int c)
{
    int ret = d_button_proc (msg, d, c);
    if (ret == D_CLOSE && d->dp3) {
        d->flags ^= D_DIRTY;
        return ((int (*)(void))d->dp3)();
    }
    return ret;
}



/* A custom dialog procedure, derived from d_button_proc. It intercepts
 * the D_CLOSE return of d_button_proc, and calls the function in dp3.
 */
int my_button_proc_ex(int msg, DIALOG *d, int c)
{
    int ret = d_button_proc (msg, d, c);
    if (ret == D_CLOSE && d->dp3) {
        d->flags ^= D_DIRTY;
        return ((int (*)(DIALOG *))d->dp3)(d);
    }
    return ret;
}



void init_dialog_colors(DIALOG *dlg)
{                  
    /* set up colors */
    gui_fg_color = makecol(0, 0, 0);
    gui_mg_color = makecol(128, 128, 128);
    gui_bg_color = makecol(224, 224, 224);
    set_dialog_color (dlg, gui_fg_color, gui_bg_color);

    /* white color for d_clear_proc and all the other procs */
    dlg[0].bg = makecol(255, 255, 255);
    for (int i = 1; dlg[i].proc; i++)
        if (dlg[i].proc == d_text_proc ||
            dlg[i].proc == d_ctext_proc ||
            dlg[i].proc == d_rtext_proc)
            dlg[i].bg = dlg[0].bg;             
}



void render_waveform(BITMAP *bmp, unsigned short *data, unsigned length, unsigned fg, unsigned bg) {
    clear_to_color(bmp, bg);
    if (data==NULL)
        return;
        
    unsigned w = bmp->w;
    unsigned h = bmp->h;
    unsigned limit = length/2;
    unsigned j, s1, s2;
    unsigned v;
    for (int i = 0; i < w; i++) {
        s1 = xtosample(i,limit,w);
        s2 = xtosample(i+1,limit,w);
        v = 0;
        for (j = s1; j < s2; j++) {
            v += (((unsigned long)data[j])*h)>>16;
        }
        v /= (s2 - s1);
	    vline(bmp, i, h/2, v, fg);
	    vline(bmp, i, h/2, h-v, fg);
    } 
}


int d_waveform_proc(int msg, DIALOG *d, int c)
{
    char **buffer  = (char **)d->dp2;
    int *buffer_size = (int*)d->dp3;
    
    switch (msg) {
      case MSG_START:
        if (d->d1 < 10)
            d->fg = makecol(255, 0, 0); 
        else
            d->fg = makecol(0, 0, 255);
        d->dp = (void*)create_bitmap(d->w, d->h);
        render_waveform((BITMAP*)d->dp, (unsigned short*)*buffer, *buffer_size, d->fg, d->bg);
        break;
        
      case MSG_END:
        destroy_bitmap((BITMAP*)d->dp);
        break;
      
      case MSG_DRAW:
        blit((BITMAP*)d->dp, screen, 0, 0, d->x, d->y, d->w, d->h);
        vline(screen, d->x+editor_dialog[d->d1+1].d2, d->y, d->y+d->h-1, (frame_counter&1) ? makecol(0,255,0) : 0);
        break;
           
      case MSG_IDLE:
        d->flags |= D_DIRTY;
        break;
    }    
    return D_O_K;    
}



struct PLAYBACKSTATE {
    int waveform_index;
    char **buffer;
    int *buffer_size;
    AUDIOSTREAM *stream;
    int delta;
};



const int STREAMINGSIZE = 4096;
const int STREAMINGLIMIT = 8192;



/* d_playback_proc updates an audiostream in MSG_IDLE if dp->stream is not NULL
 * it is a non visual element
 * DURING INIT:
 * d1 is index of associated d_waveform_proc
 *
 * WHILE PLAYING:
 * dp is PLAYBACKSTATE
 * d1 is next sample to load
 * d2 is sample to end playback
 */
int d_playback_proc(int msg, DIALOG *d, int c)
{
    PLAYBACKSTATE* pbs = (PLAYBACKSTATE*)d->dp;
    switch (msg) {
      case MSG_START:
        pbs = (PLAYBACKSTATE*)malloc(sizeof(PLAYBACKSTATE));
        pbs->waveform_index = d->d1;
        pbs->buffer = (char**)d->dp2;//editor_dialog[d->d2].dp2;
        pbs->buffer_size = (int*)d->dp3;//editor_dialog[d->d2].dp3;
        pbs->stream = NULL;
        d->dp = (void*)pbs;
        d->d1 = 0;
        d->d2 = 0;
        break;
           
      case MSG_END:
        if (pbs->stream) stop_audio_stream(pbs->stream);
        free(pbs);
        break;
        
      case MSG_IDLE:
        if (pbs->stream != NULL) {
            unsigned short *mem_chunk = (unsigned short *)get_audio_stream_buffer(pbs->stream);
            if (mem_chunk != NULL) {  // Refill the stream buffer.
                int length = MIN(d->d2 - d->d1, STREAMINGSIZE);
                if (length > 0) { 
                    memset(mem_chunk, 127, STREAMINGLIMIT);
                    //d->d1 += resample_buffer_up(mem_chunk, ((unsigned short*)*pbs->buffer)+d->d1, STREAMINGLIMIT, length*2, 11025, 8000, pbs->delta);
                    memcpy(mem_chunk, ((unsigned short*)*pbs->buffer)+d->d1, length*2);
                    free_audio_stream_buffer(pbs->stream);
                    d->d1 += STREAMINGSIZE;
                    if (d->d1 > d->d2 ) { d->d1 = d->d2; }
                    editor_dialog[editor_dialog[pbs->waveform_index].d1+1].d2 = // playback position control
                        xtosample(d->d1, editor_dialog[pbs->waveform_index].w, *pbs->buffer_size/2); // sample to position
                     editor_dialog[editor_dialog[pbs->waveform_index].d1+1].flags |= D_DIRTY;
                } else {
                    memset(mem_chunk, 127, STREAMINGSIZE);
                    free_audio_stream_buffer(pbs->stream);
                    stop_audio_stream(pbs->stream);
                    pbs->stream = NULL;
                }
                   
            }
        }
        break;
    }
    return D_O_K;
}



int play(DIALOG *d)
{
    PLAYBACKSTATE* pbs = (PLAYBACKSTATE*)editor_dialog[d->d1].dp;
    
    if ( pbs->stream == NULL) {
        unsigned sela = editor_dialog[editor_dialog[pbs->waveform_index].d1+1].d2;
        int *buffer_size = (int*)editor_dialog[pbs->waveform_index].dp3;
        unsigned sample_begin = xtosample(sela, *buffer_size/2, editor_dialog[2].w);
        unsigned sample_end = *buffer_size/2;;
        editor_dialog[d->d1].d1 = sample_begin;
        editor_dialog[d->d1].d2 = sample_end;
                        
        pbs->delta = 0;
        pbs->stream = play_audio_stream(STREAMINGSIZE, 16, FALSE, 11025, 255, 128);
        if (pbs->stream == NULL) // some sort of internal allegro error
           alert("Playback", "", "Couldn't begin", "Ok I'll try again", NULL, 0, 0);
    }
    return D_O_K;    
}



int pause(DIALOG *d)
{
    PLAYBACKSTATE* pbs = (PLAYBACKSTATE*)editor_dialog[d->d1].dp;
    if ( pbs->stream != NULL) {
        stop_audio_stream(pbs->stream);
        pbs->stream = NULL;
    }
    else
        pbs->stream = play_audio_stream(STREAMINGSIZE, 16, FALSE, 11025, 255, 128);        
    return D_O_K;    
}



int stop(DIALOG *d)
{
    PLAYBACKSTATE* pbs = (PLAYBACKSTATE*)editor_dialog[d->d1].dp;
    if ( pbs->stream != NULL ) {
        stop_audio_stream(pbs->stream);
        pbs->stream = NULL;
    }
    return D_O_K;    
}



void player_window(void)
{
	init_dialog_colors(editor_dialog);
    do_dialog(editor_dialog, -1);
    
    if (buffer) { free(buffer); }
}
