#include <allegro.h>
#include <mem.h>
#include "sndhelp.h"
#include "editor.h"

char* buffer = NULL;
int buffer_size = 0;

int start(void);
int stop(void);
int done(void);
int cancel(void);
int d_capture_proc(int msg, DIALOG *d, int c);

bool capturing;               // if we are capturing data from the sound card
int scratch_buffer_size;      // size of the intermediate scratch buffer
unsigned short *scratch_buffer; // intermediate scratch buffer where data goes before proecessing

char samples_as_seconds[20];  // used by field 3 -- seconds
char samples_as_text[20];     // used by field 5 -- sample size of buffer

DIALOG capture_dialog[] =
{
    /* (dialog proc)     (x)   (y)   (w)   (h) (fg)(bg) (key) (flags)     (d1) (d2)    (dp)                   (dp2) (dp3) */
   
    /* this element is a shadow box therefore it should come before the others */
    { d_shadow_box_proc,    0,   0,  600,  200,   0,  0,   0,      0,       0,   0,    NULL,                   NULL, NULL  },  
    { d_ctext_proc,        10,  10,  580,  180,   0,  0,   0,      0,       0,   0,    (void*)"Capture waveform", NULL, NULL },

    /* lots of descriptive text elements */
    { d_text_proc,         20,  50,   0,    0,   0,  0,    0,      0,       0,   0,    (void*)"Seconds captured:", NULL, NULL },
    { d_rtext_proc,       240,  50,   0,    0,   0,  0,    0,      0,       0,   0,    (void*)"00:00" ,        NULL, NULL },
    { d_text_proc,         20,  70,   0,    0,   0,  0,    0,      0,       0,   0,    (void*)"Size of buffer:", NULL, NULL },
    { d_rtext_proc,       240,  70,   0,    0,   0,  0,    0,      0,       0,   0,    (void*)"00000000000",  NULL, NULL },
    
    /* the cancel buttons and start/stop use our customized dialog procedure, using dp3 as callback */
    { my_button_proc,    430, 50,  160,   20,   0,  0,  's', D_EXIT,       0,   0,    (void*)"&Start",        NULL, (void*)start },
    { my_button_proc,    430, 80,  160,   20,   0,  0,  't', D_DISABLED|D_EXIT,       0,   0,    (void*)"S&top",         NULL, (void*)stop },
    { my_button_proc,    430, 140,  160,   20,   0,  0,  'd',  D_EXIT,       0,   0,    (void*)"&Done",         NULL, (void*)done },
    { my_button_proc,    430, 170,  160,   20,   0,  0,  'c', D_EXIT,       0,   0,    (void*)"&Cancel",       NULL, (void *)cancel },
    
    /* the record proc consumes all cpu cycles ^_^ */
    { d_capture_proc,     20,  90,   390,   90,  0,  0,    0,      0,       0,   0,    NULL,                   NULL, NULL  },
    //{ d_yield_proc,        0,   0,    0,    0,   0,  0,    0,      0,       0,   0,    NULL,                   NULL, NULL  },
    { NULL,                0,   0,    0,    0,   0,  0,    0,      0,       0,   0,    NULL,                   NULL, NULL  }
};


/* Used by the start button. */
int start(void)
{
    capture_dialog[6].flags |= D_DISABLED;
    capture_dialog[7].flags &= ~D_DISABLED;
    capture_dialog[7].flags |= D_DIRTY;
    capturing = true;
    return D_O_K;
}



/* Used by the stop button. */
int stop(void)
{
    capturing = false;
    capture_dialog[7].flags |= D_DISABLED;
    capture_dialog[6].flags &= ~D_DISABLED;
    capture_dialog[6].flags |= D_DIRTY;
    return D_O_K;
}



/* Used by the done button. */
int done(void)
{
    if (capturing) {
        if (alert("Paused", NULL, "Are you sure you want to stop?", "&Yes", "&No", 'y', 'n') == 1)
            return D_CLOSE;
        else
            return D_O_K;
    }
    return D_CLOSE;
}



/* Used by the cancel button. */
int cancel(void)
{
    if (capturing) {
        if (alert("Paused", NULL, "Are you sure you want to stop?", "&Yes", "&No", 'y', 'n') == 1)
            return D_CLOSE;
        else
            return D_O_K;
    } else {
        return D_CLOSE;
    }
}



/* Used to update the display of samples captured */
void update_display(void) {
    //int sec = buffer_size / (8000 * 2);
    int sec = buffer_size / (11025 * 2);
    int min = sec / 60;
    usprintf(samples_as_seconds, "%02d:%02d", min, sec%60);
    usprintf(samples_as_text, "%dk", buffer_size/1024);
    
    capture_dialog[3].dp = samples_as_seconds;
    capture_dialog[5].dp = samples_as_text;
    capture_dialog[3].flags |= D_DIRTY;
    capture_dialog[5].flags |= D_DIRTY;
}



/* Used to capture data */
int d_capture_proc(int msg, DIALOG *d, int c)
{
    switch (msg) {
      case MSG_START:
        d->bg = makecol(255,255,255);
        d->fg = makecol(255,0,0);
        update_display();
        scratch_buffer_size = start_sound_input(11025, 16, 0);
        if (scratch_buffer_size > 0) {
            scratch_buffer = (unsigned short*)malloc(scratch_buffer_size);
        } else {
            allegro_message("sound card not capable of 8khz 16bit capture");
            return D_CLOSE;
        }
        d->dp = (void*)create_bitmap(d->w, d->h);
        break;
        
      case MSG_END:
        stop_sound_input();
        free(scratch_buffer);       // no need to check for null as it is only possible
        scratch_buffer = 0;         // to be here if scratchbuffer is allocated
        destroy_bitmap((BITMAP*)d->dp);
        break;
        
      case MSG_DRAW:
        {
            int w = MIN(d->w, scratch_buffer_size/2);
            int y, bg = d->bg, fg = d->fg;
            
      	    for (int i=0; i<w; i++) {
    	       if (scratch_buffer[i] < 0x8000) {
                  y = scratch_buffer[i]*d->h/65535;
        		  vline((BITMAP*)d->dp, i, 0, y-1, bg);
        		  vline((BITMAP*)d->dp, i, y, d->h/2, fg);
        		  vline((BITMAP*)d->dp, i, d->h/2+1, d->h, bg);
    	       }
    	       else {
                  y = scratch_buffer[i]*d->h/65535;
        		  vline((BITMAP*)d->dp, i, 0, d->h/2-1, bg);
        		  vline((BITMAP*)d->dp, i, d->h/2,y, fg);
        		  vline((BITMAP*)d->dp, i, y+1, d->h, bg);
    	       }
    	    }
    	    blit((BITMAP*)d->dp, screen, 0, 0, d->x, d->y, d->w, d->h);
        }
	    break;
      
    
      case MSG_IDLE:           
        if (read_sound_input(scratch_buffer) && capturing) {
                if (!buffer && buffer_size == 0) // fix up realloc of NULL bug
                    buffer = (char*)malloc(scratch_buffer_size);
                else
                    buffer = (char*)realloc(buffer, buffer_size + scratch_buffer_size);
                memcpy(buffer+buffer_size, scratch_buffer, scratch_buffer_size);
                buffer_size += scratch_buffer_size; 
            update_display();
        }
        d->flags |= D_DIRTY;
        break;
    }
    return D_O_K;
}



void capture_window(void)
{
    int start_buffer_size = buffer_size;
     
    capturing = false;
    scratch_buffer_size = 0;
    
    centre_dialog(capture_dialog);
    init_dialog_colors(capture_dialog);
    int ret = popup_dialog(capture_dialog, 6);
    if (ret == -1 || ret == 9) {
        if (buffer) {
            if (start_buffer_size == 0) {
                free(buffer);
                buffer = NULL;
                buffer_size = 0;
            } else {
                buffer = (char*)realloc(buffer, start_buffer_size);
                buffer_size = (buffer != NULL) ? start_buffer_size : 0;
            }
        }
    }
}
