#include <allegro.h>
#include <speex/speex.h>
#include <math.h>
#include <mem.h>
#include <stdio.h>
#include "sndhelp.h"


void play_test_tone(void)
{
    int dwLength = 36000;
    float  amp[] = {8000.0f, 8000.0f, 8000.0f, 8000.0f};
    unsigned begin[] = {0, 5000, 10000, 15000};
    float  factor[] = {M_PI/60.0, M_PI/50.0, M_PI/40.0, M_PI/30.0};
    SAMPLE *smp = create_sample(16, 0, 44100, dwLength);
    if (smp == NULL) { allegro_message("couldn't create sample buffer!"); return; } 
    short *lpWrite = (short*)smp->data;  // Sound card data is unsigned but
                                         // calculating signed data is easier
    for (unsigned int i = 0; i < dwLength; i ++) {
                              lpWrite[i]  = (short)(amp[0]*sin(i*factor[0])); amp[0] *= 0.9997;
            if (i>begin[1]) { lpWrite[i] += (short)(amp[1]*sin(i*factor[1])); amp[1] *= 0.9997; }
            if (i>begin[2]) { lpWrite[i] += (short)(amp[2]*sin(i*factor[2])); amp[2] *= 0.9997; }
            if (i>begin[3]) { lpWrite[i] += (short)(amp[3]*sin(i*factor[3])); amp[3] *= 0.9997; }
            ((unsigned short*)lpWrite)[i] ^= 0x8000;
    }
    int voice = play_sample(smp, 255, 128, 1000, 0);
    if (voice < 0) { allegro_message("no voices available!"); }
    while (voice_get_position(voice) >= 0);
    destroy_sample(smp);
}



#define FRAME_SIZE 160
void write_buffer_to_file(char *filename, char *buffer, int buffer_size)
{
    // buffer and buffer_size are length of the sample buffer in bytes,
    // not samples, so first we have to convert before doing math on it.
    // We pass arrays around this way to prevent allocation confusion.
    unsigned short *short_buffer = (unsigned short*)buffer;
    int short_buffer_size = buffer_size / 2;

    // speex manual says we should resample from 11025 Hz to 8000 Hz
    // before saving, but the codec seems to handle it.
    float input[FRAME_SIZE];
    char data[256];
    int numbytes;
    void *enc_state;
    SpeexBits bits;
    int pos = 0;
    int tmp = 8; // (15 kbps)
    int len;
    FILE *fp = fopen(filename, "wb");
    if (!fp) { alert("Error", "", "Couldn't write to file or file in use by another application", "Ok", NULL, 0, 0); return; }
    enc_state = speex_encoder_init(&speex_nb_mode);
    speex_encoder_ctl(enc_state, SPEEX_SET_QUALITY, &tmp);
    speex_bits_init(&bits);
    while (true) {
        len = MIN(FRAME_SIZE, short_buffer_size - pos);
        if (len < FRAME_SIZE) { break; } // chuck out last frame
        // copy 16-bit values to float so speex can work on them
        for (int i = 0; i < len; i++)
            input[i] = short_buffer[pos + i] - 0x8000; 
        pos += len;
        // flush the bits the struct to work on a new set
        speex_bits_reset(&bits);
        // encode the frame
        speex_encode(enc_state, input, &bits);
        numbytes = speex_bits_write(&bits, data, 256);
        fwrite(&numbytes, sizeof(int), 1, fp);
        fwrite(data, 1, numbytes, fp);
    }
    // clean up
    speex_encoder_destroy(enc_state);
    speex_bits_destroy(&bits);
    fclose(fp);                         
}


void read_buffer_from_file(char *filename, char* &buffer, int &buffer_size)
{
    // buffer and buffer_size are length of the sample buffer in bytes,
    // not samples, so first we have to convert before doing math on it.
    // We pass arrays around this way to prevent allocation confusion.
    short *short_buffer = (short*)buffer;
    int short_buffer_size = buffer_size / 2;

    float output[FRAME_SIZE];
    char data[200];
    int numbytes;
    void *dec_state;
    SpeexBits bits;
    FILE *fp = fopen(filename, "rb");
    int tmp = 1;
    if (!fp) { alert("Error", "", "Couldn't open file or file in use by another application", "Ok", NULL, 0, 0); return; }
    buffer_size = 0;
    dec_state = speex_decoder_init(&speex_nb_mode);
    speex_decoder_ctl(dec_state, SPEEX_SET_ENH, &tmp);
    speex_bits_init(&bits);
    if (buffer == NULL) {
        short_buffer = (short*)malloc(4); // fix up realloc of NULL bug
        buffer = (char*)short_buffer;         
    }
    while (true) {
        fread(&numbytes, sizeof(int), 1, fp);
        if (feof(fp)) { break; }
        fread(data, 1, numbytes, fp);
        // decoder frame
        speex_bits_read_from(&bits, data, numbytes);
        speex_decode(dec_state, &bits, output);
        // output to buffer
        short_buffer = (short*)realloc(buffer, buffer_size + (FRAME_SIZE*2));
        buffer = (char*)short_buffer;
        for (int i=0; i<FRAME_SIZE; i++) {
           short_buffer[buffer_size/2 + i] = output[i];
           short_buffer[buffer_size/2 + i] ^= 0x8000;
        }
        buffer_size += FRAME_SIZE*2;
    }
    speex_encoder_destroy(dec_state);
    speex_bits_destroy(&bits);
    fclose(fp);
    

}
